"""Tests for Chaos Engineering middleware."""

import time

import pytest
from httpx import ASGITransport, AsyncClient

from llmock.chaos import ChaosSettings
from llmock.main import create_app


@pytest.fixture
def app():
    return create_app(chaos=ChaosSettings())


async def test_no_chaos_passes_through(app):
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/v1/models")
    assert resp.status_code == 200


async def test_health_bypasses_chaos(app):
    """Health endpoint must always succeed regardless of error rates."""
    app.state.chaos_settings.error_rate_429 = 0.6
    app.state.chaos_settings.error_rate_500 = 0.4
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/health")
    assert resp.status_code == 200


async def test_429_always_injected(app):
    app.state.chaos_settings.error_rate_429 = 1.0
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/v1/models")
    assert resp.status_code == 429
    body = resp.json()
    assert body["error"]["code"] == "rate_limit_exceeded"


async def test_500_always_injected(app):
    app.state.chaos_settings.error_rate_500 = 1.0
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/v1/models")
    assert resp.status_code == 500
    body = resp.json()
    assert body["error"]["code"] == "internal_error"


async def test_503_always_injected(app):
    app.state.chaos_settings.error_rate_503 = 1.0
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/v1/models")
    assert resp.status_code == 503
    body = resp.json()
    assert body["error"]["code"] == "service_unavailable"


async def test_arbitrary_4xx_error_is_injected_with_provider_shape(app):
    app.state.chaos_settings.error_rate_451 = 1.0
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/v1/models")
    assert resp.status_code == 451
    body = resp.json()
    assert body["error"]["type"] == "invalid_request_error"
    assert body["error"]["code"] == "http_451"


async def test_arbitrary_5xx_error_is_injected_with_gemini_shape(app):
    app.state.chaos_settings.error_rate_507 = 1.0
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/gemini/v1beta/models")
    assert resp.status_code == 507
    body = resp.json()
    assert body["error"]["status"] == "INTERNAL"
    assert body["error"]["code"] == 507


async def test_multiple_error_rates_are_sampled(app, monkeypatch):
    app.state.chaos_settings.error_rate_429 = 0.6
    app.state.chaos_settings.error_rate_503 = 0.4
    monkeypatch.setattr("llmock.chaos.random.random", lambda: 0.7)
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        resp = await client.get("/v1/models")
    assert resp.status_code == 503


async def test_latency_injected(app):
    app.state.chaos_settings.latency_ms = 200
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        start = time.monotonic()
        resp = await client.get("/v1/models")
        elapsed_ms = (time.monotonic() - start) * 1000
    assert resp.status_code == 200
    assert elapsed_ms >= 190  # allow 10ms tolerance


async def test_zero_error_rate_never_injects(app):
    """With all rates at 0, no error should ever be injected over many requests."""
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        for _ in range(20):
            resp = await client.get("/v1/models")
            assert resp.status_code == 200


def test_chaos_settings_from_env(monkeypatch):
    monkeypatch.setenv("LLMOCK_LATENCY_MS", "150")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_401", "0.2")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_429", "0.3")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_451", "0.15")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_500", "0.1")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_503", "0.05")
    cfg = ChaosSettings.from_env()
    assert cfg.latency_ms == 150
    assert cfg.error_rate_401 == 0.2
    assert cfg.error_rate_429 == 0.3
    assert cfg.error_rate_451 == 0.15
    assert cfg.error_rate_500 == 0.1
    assert cfg.error_rate_503 == 0.05


def test_create_app_uses_explicit_chaos_settings():
    cfg = ChaosSettings(latency_ms=500, error_rate_429=0.5)
    app = create_app(chaos=cfg)
    assert app.state.chaos_settings is cfg
    assert app.state.chaos_settings.latency_ms == 500
    assert app.state.chaos_settings.error_rate_429 == 0.5


def test_create_app_reads_chaos_settings_from_env(monkeypatch):
    monkeypatch.setenv("LLMOCK_LATENCY_MS", "250")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_401", "0.25")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_429", "0.4")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_507", "0.05")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_500", "0.2")
    monkeypatch.setenv("LLMOCK_ERROR_RATE_503", "0.05")
    app = create_app()
    cfg = app.state.chaos_settings
    assert cfg.latency_ms == 250
    assert cfg.error_rate_401 == 0.25
    assert cfg.error_rate_429 == 0.4
    assert cfg.error_rate_507 == 0.05
    assert cfg.error_rate_500 == 0.2
    assert cfg.error_rate_503 == 0.05


def test_invalid_total_error_probability_raises():
    with pytest.raises(ValueError):
        ChaosSettings(error_rate_429=0.8, error_rate_503=0.5).validated()


def test_chaos_settings_accept_arbitrary_status_kwargs():
    cfg = ChaosSettings(error_rate_401=0.2, error_rate_504=0.1)
    assert cfg.error_rate_401 == 0.2
    assert cfg.error_rate_504 == 0.1
    assert cfg.error_rates[401] == 0.2
    assert cfg.error_rates[504] == 0.1


def test_chaos_settings_support_dynamic_status_assignment():
    cfg = ChaosSettings()
    cfg.error_rate_422 = 0.35
    assert cfg.error_rate_422 == 0.35
    assert cfg.error_rates[422] == 0.35


def test_invalid_http_error_code_raises():
    with pytest.raises(ValueError):
        ChaosSettings(error_rate_200=0.2).validated()



"""Tests for configurable success payloads and provider-specific error shapes."""

from fastapi.testclient import TestClient

from llmock.chaos import ChaosSettings
from llmock.main import create_app
from llmock.simulation import MockResponseSettings


def test_response_style_hello_returns_greeting():
    client = TestClient(create_app(responses=MockResponseSettings(response_style="hello")))

    resp = client.post(
        "/v1/chat/completions",
        json={
            "model": "gpt-4o",
            "messages": [{"role": "user", "content": "Tell me something"}],
        },
    )

    assert resp.status_code == 200
    assert resp.json()["choices"][0]["message"]["content"].startswith("Hello!")


def test_response_style_echo_reflects_prompt():
    client = TestClient(create_app(responses=MockResponseSettings(response_style="echo")))

    resp = client.post(
        "/anthropic/v1/messages",
        json={
            "model": "claude-sonnet-4-6",
            "max_tokens": 128,
            "messages": [{"role": "user", "content": "Echo this mock prompt"}],
        },
    )

    assert resp.status_code == 200
    text = resp.json()["content"][0]["text"]
    assert "Echo this mock prompt" in text


def test_force_status_header_uses_gemini_error_shape():
    client = TestClient(create_app())

    resp = client.get(
        "/gemini/v1beta/models",
        headers={"x-llmock-force-status": "429"},
    )

    assert resp.status_code == 429
    body = resp.json()
    assert body["error"]["status"] == "RESOURCE_EXHAUSTED"
    assert body["error"]["code"] == 429


def test_force_status_header_uses_cohere_error_shape():
    client = TestClient(create_app())

    resp = client.post(
        "/cohere/v2/chat",
        json={"model": "command-r", "messages": [{"role": "user", "content": "Hi"}]},
        headers={"x-llmock-force-status": "401"},
    )

    assert resp.status_code == 401
    body = resp.json()
    assert body["type"] == "authentication_error"
    assert body["message"]


def test_anthropic_supports_provider_specific_529_overload():
    client = TestClient(create_app(chaos=ChaosSettings(error_rates={529: 1.0})))

    resp = client.post(
        "/anthropic/v1/messages",
        json={
            "model": "claude-opus-4-6",
            "max_tokens": 64,
            "messages": [{"role": "user", "content": "Hello"}],
        },
    )

    assert resp.status_code == 529
    body = resp.json()
    assert body["type"] == "error"
    assert body["error"]["type"] == "overloaded_error"



"""Shared simulation helpers for LLMock routers and middleware."""

from __future__ import annotations

import base64
import hashlib
import html
from http import HTTPStatus
import os
import random
from dataclasses import dataclass
from typing import Any

from starlette.responses import JSONResponse

SUPPORTED_RESPONSE_STYLES = {"static", "hello", "echo", "varied"}
DEFAULT_IMAGE_SIZE = "1024x1024"
ERROR_RATE_ENV_PREFIX = "LLMOCK_ERROR_RATE_"
SUPPORTED_ERROR_STATUS_CODES = tuple(range(400, 600))


def raise_if_streaming(stream: bool) -> None:
    """Raise HTTPException 501 when a client requests streaming mode.

    LLMock does not implement SSE streaming. Returning 501 instead of silently
    ignoring the flag prevents SDK streaming iterators from hanging forever.
    """
    if stream:
        from fastapi import HTTPException  # local import avoids circularity
        raise HTTPException(status_code=501, detail="Streaming is not supported by LLMock. Set stream=false.")


@dataclass
class MockResponseSettings:
    response_style: str = "varied"

    @classmethod
    def from_env(cls) -> "MockResponseSettings":
        return cls(
            response_style=os.getenv("LLMOCK_RESPONSE_STYLE", "varied"),
        ).validated()

    def validated(self) -> "MockResponseSettings":
        if self.response_style not in SUPPORTED_RESPONSE_STYLES:
            supported = ", ".join(sorted(SUPPORTED_RESPONSE_STYLES))
            raise ValueError(f"LLMOCK_RESPONSE_STYLE must be one of: {supported}.")
        return self

    def as_env(self) -> dict[str, str]:
        return {"LLMOCK_RESPONSE_STYLE": self.response_style}


def flatten_text(value: Any) -> str:
    """Flatten nested provider content blocks into a readable text string."""
    if isinstance(value, str):
        return value
    if isinstance(value, list):
        return " ".join(part for part in (flatten_text(item) for item in value) if part)
    if isinstance(value, dict):
        if isinstance(value.get("text"), str):
            return value["text"]
        if "inline_data" in value or "image_url" in value or value.get("type") == "image_url":
            return "[image input]"
        for key in ("content", "parts", "input", "message"):
            if key in value:
                return flatten_text(value[key])
    return ""


def estimate_tokens(*values: Any) -> int:
    return sum(len(flatten_text(value)) // 4 for value in values)


def build_mock_text(
    *,
    settings: MockResponseSettings,
    model: str,
    prompt: str = "",
) -> str:
    prompt = " ".join(prompt.split()).strip()
    if settings.response_style == "static":
        return f"Mock response from {model}."
    if settings.response_style == "hello":
        return f"Hello! This is a mock response from {model}."
    if settings.response_style == "echo":
        if prompt:
            return f"Hello! You said: {prompt[:120]}"
        return f"Hello! This is a mock response from {model}."

    templates = [
        "Hello! LLMock handled this request for {model}.",
        "Mock response from {model}: everything is working as expected.",
        "Hi there. This is a simulated reply from {model}.",
        "LLMock says hello from {model}.",
        "Simulated success from {model}.",
    ]
    if prompt:
        templates.extend(
            [
                "Hello! {model} received: {prompt}",
                "{model} mock reply: I saw '{prompt}'.",
                "Simulated answer from {model} for '{prompt}'.",
            ]
        )

    digest = hashlib.sha256(f"{model}|{prompt}".encode("utf-8")).digest()
    template = templates[digest[0] % len(templates)]
    prompt_snippet = prompt[:80] if prompt else "your prompt"
    return template.format(model=model, prompt=prompt_snippet)


def build_mock_embedding(text: str, dimension: int = 1536) -> list[float]:
    seed = int.from_bytes(hashlib.sha256(text.encode("utf-8")).digest()[:8], "big")
    rng = random.Random(seed)
    return [round(rng.uniform(-1.0, 1.0), 6) for _ in range(dimension)]


def _build_svg(prompt: str, size: str) -> str:
    width_text, _, height_text = size.partition("x")
    width = int(width_text) if width_text.isdigit() else 1024
    height = int(height_text) if height_text.isdigit() else 1024
    safe_prompt = html.escape(prompt[:64] or "LLMock image")
    return (
        f"<svg xmlns='http://www.w3.org/2000/svg' width='{width}' height='{height}' "
        f"viewBox='0 0 {width} {height}'>"
        "<defs><linearGradient id='g' x1='0' x2='1' y1='0' y2='1'>"
        "<stop offset='0%' stop-color='#111827'/>"
        "<stop offset='100%' stop-color='#2563eb'/>"
        "</linearGradient></defs>"
        f"<rect width='{width}' height='{height}' fill='url(#g)'/>"
        f"<text x='50%' y='48%' dominant-baseline='middle' text-anchor='middle' "
        "font-family='monospace' font-size='42' fill='white'>LLMock</text>"
        f"<text x='50%' y='58%' dominant-baseline='middle' text-anchor='middle' "
        "font-family='monospace' font-size='24' fill='#bfdbfe'>"
        f"{safe_prompt}</text>"
        "</svg>"
    )


def build_fake_image_data_uri(prompt: str, size: str = DEFAULT_IMAGE_SIZE) -> str:
    svg = _build_svg(prompt, size)
    payload = base64.b64encode(svg.encode("utf-8")).decode("ascii")
    return f"data:image/svg+xml;base64,{payload}"


def build_fake_image_payload(
    *,
    prompt: str,
    count: int,
    size: str = DEFAULT_IMAGE_SIZE,
    response_format: str = "url",
) -> list[dict[str, str]]:
    data_uri = build_fake_image_data_uri(prompt=prompt, size=size)
    if response_format == "b64_json":
        return [
            {"b64_json": data_uri.split(",", 1)[1], "revised_prompt": prompt}
            for _ in range(count)
        ]
    return [{"url": data_uri, "revised_prompt": prompt} for _ in range(count)]


def parse_error_rate_env() -> dict[int, float]:
    rates: dict[int, float] = {}
    for key, value in os.environ.items():
        if not key.startswith(ERROR_RATE_ENV_PREFIX):
            continue
        suffix = key[len(ERROR_RATE_ENV_PREFIX) :]
        if not suffix.isdigit():
            continue
        rates[int(suffix)] = float(value)
    return rates


def provider_from_path(path: str) -> str:
    if path.startswith("/anthropic/"):
        return "anthropic"
    if path.startswith("/gemini/"):
        return "gemini"
    if path.startswith("/cohere/"):
        return "cohere"
    if path.startswith("/mistral/"):
        return "mistral"
    if path.startswith("/groq/"):
        return "groq"
    if path.startswith("/together/"):
        return "together"
    if path.startswith("/perplexity/"):
        return "perplexity"
    if path.startswith("/ai21/"):
        return "ai21"
    if path.startswith("/xai/"):
        return "xai"
    return "openai"


def build_error_response(path: str, status_code: int) -> JSONResponse:
    provider = provider_from_path(path)
    content = _build_error_content(provider=provider, status_code=status_code)
    headers = {"retry-after": "1"} if status_code in {429, 503, 504, 529} else None
    return JSONResponse(status_code=status_code, content=content, headers=headers)


def _build_error_content(provider: str, status_code: int) -> dict[str, Any]:
    label = {
        400: "Bad request.",
        401: "Unauthorized.",
        402: "Payment required.",
        403: "Forbidden.",
        404: "Resource not found.",
        408: "Request timeout.",
        409: "Conflict.",
        413: "Payload too large.",
        422: "Unprocessable entity.",
        429: "Rate limit exceeded.",
        500: "Internal server error.",
        501: "Not implemented.",
        502: "Bad gateway.",
        503: "Service unavailable.",
        504: "Gateway timeout.",
        529: "Service overloaded.",
    }.get(status_code, _http_status_label(status_code))

    openai_types = {
        400: ("invalid_request_error", "bad_request"),
        401: ("authentication_error", "invalid_api_key"),
        402: ("billing_error", "billing_required"),
        403: ("permission_error", "forbidden"),
        404: ("invalid_request_error", "not_found"),
        408: ("request_timeout", "request_timeout"),
        409: ("conflict_error", "conflict"),
        413: ("invalid_request_error", "request_too_large"),
        422: ("invalid_request_error", "unprocessable_entity"),
        429: ("rate_limit_error", "rate_limit_exceeded"),
        500: ("server_error", "internal_error"),
        501: ("server_error", "not_implemented"),
        502: ("server_error", "bad_gateway"),
        503: ("server_error", "service_unavailable"),
        504: ("server_error", "gateway_timeout"),
        529: ("server_error", "overloaded"),
    }

    if provider == "anthropic":
        anthropic_type = {
            400: "invalid_request_error",
            401: "authentication_error",
            402: "billing_error",
            403: "permission_error",
            404: "not_found_error",
            408: "api_error",
            409: "api_error",
            413: "request_too_large",
            422: "invalid_request_error",
            429: "rate_limit_error",
            500: "api_error",
            501: "api_error",
            502: "api_error",
            503: "overloaded_error",
            504: "api_error",
            529: "overloaded_error",
        }.get(status_code, _fallback_anthropic_error_type(status_code))
        return {"type": "error", "error": {"type": anthropic_type, "message": label}}

    if provider == "gemini":
        gemini_status = {
            400: "INVALID_ARGUMENT",
            401: "UNAUTHENTICATED",
            402: "FAILED_PRECONDITION",
            403: "PERMISSION_DENIED",
            404: "NOT_FOUND",
            408: "DEADLINE_EXCEEDED",
            409: "ABORTED",
            413: "RESOURCE_EXHAUSTED",
            422: "FAILED_PRECONDITION",
            429: "RESOURCE_EXHAUSTED",
            500: "INTERNAL",
            501: "UNIMPLEMENTED",
            502: "UNAVAILABLE",
            503: "UNAVAILABLE",
            504: "DEADLINE_EXCEEDED",
            529: "UNAVAILABLE",
        }.get(status_code, _fallback_gemini_status(status_code))
        return {"error": {"code": status_code, "message": label, "status": gemini_status}}

    if provider == "cohere":
        cohere_type = {
            400: "invalid_request_error",
            401: "authentication_error",
            402: "billing_error",
            403: "permission_error",
            404: "not_found_error",
            408: "timeout_error",
            409: "conflict_error",
            413: "request_too_large",
            422: "unprocessable_entity_error",
            429: "rate_limit_error",
            500: "server_error",
            501: "server_error",
            502: "server_error",
            503: "server_error",
            504: "server_error",
            529: "server_error",
        }.get(status_code, _fallback_cohere_error_type(status_code))
        return {"message": label, "type": cohere_type}

    if provider == "mistral":
        error_type, error_code = openai_types.get(status_code, _fallback_openai_error(status_code))
        return {
            "object": "error",
            "message": label,
            "type": error_type,
            "param": None,
            "code": error_code,
        }

    error_type, error_code = openai_types.get(status_code, _fallback_openai_error(status_code))
    return {
        "error": {
            "message": label,
            "type": error_type,
            "param": None,
            "code": error_code,
        }
    }


def _http_status_label(status_code: int) -> str:
    try:
        phrase = HTTPStatus(status_code).phrase
    except ValueError:
        phrase = f"HTTP {status_code} error"
    return f"{phrase}."


def _fallback_openai_error(status_code: int) -> tuple[str, str]:
    if 400 <= status_code < 500:
        return ("invalid_request_error", f"http_{status_code}")
    return ("server_error", f"http_{status_code}")


def _fallback_anthropic_error_type(status_code: int) -> str:
    if 400 <= status_code < 500:
        return "invalid_request_error"
    return "api_error"


def _fallback_gemini_status(status_code: int) -> str:
    if 400 <= status_code < 500:
        return "INVALID_ARGUMENT"
    return "INTERNAL"


def _fallback_cohere_error_type(status_code: int) -> str:
    if 400 <= status_code < 500:
        return "invalid_request_error"
    return "server_error"



"""Provider-specific exception handlers for LLMock.

Overrides FastAPI's default `{"detail": "..."}` shape with the real
error envelope each provider uses so that clients can parse them normally.
"""

from __future__ import annotations

from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException
from starlette.responses import JSONResponse

from llmock.simulation import _build_error_content, provider_from_path


def _message_from_exc(exc: HTTPException) -> str:
    detail = exc.detail
    if isinstance(detail, str):
        return detail
    if isinstance(detail, dict):
        return str(detail.get("msg") or detail.get("message") or exc.detail)
    return str(detail)


def _message_from_validation(exc: RequestValidationError) -> str:
    errors = exc.errors()
    if not errors:
        return "Request validation error."
    first = errors[0]
    loc = " → ".join(str(p) for p in first.get("loc", []))
    msg = first.get("msg", "validation error")
    return f"{loc}: {msg}" if loc else msg


async def _http_exception_handler(request: Request, exc: HTTPException) -> JSONResponse:
    provider = provider_from_path(request.url.path)
    message = _message_from_exc(exc)
    content = _build_error_content(provider=provider, status_code=exc.status_code)

    # Inject the actual message instead of the generic label.
    _inject_message(content, message, provider)

    headers = dict(exc.headers or {})
    if exc.status_code in {429, 503, 504, 529} and "retry-after" not in {k.lower() for k in headers}:
        headers["retry-after"] = "1"

    return JSONResponse(status_code=exc.status_code, content=content, headers=headers or None)


async def _validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    provider = provider_from_path(request.url.path)
    status_code = 422
    message = _message_from_validation(exc)
    content = _build_error_content(provider=provider, status_code=status_code)
    _inject_message(content, message, provider)
    return JSONResponse(status_code=status_code, content=content)


def _inject_message(content: dict, message: str, provider: str) -> None:
    """Mutate *content* in-place to use *message* instead of the generic label."""
    if provider == "anthropic":
        # {"type": "error", "error": {"type": "...", "message": "..."}}
        if isinstance(content.get("error"), dict):
            content["error"]["message"] = message
    elif provider == "gemini":
        # {"error": {"code": ..., "message": "...", "status": "..."}}
        if isinstance(content.get("error"), dict):
            content["error"]["message"] = message
    elif provider == "cohere":
        # {"message": "..."}
        content["message"] = message
    elif provider == "mistral":
        # {"object": "error", "message": "...", "type": "...", "param": null, "code": "..."}
        content["message"] = message
    elif provider == "xai":
        # {"error": {"message": "...", "type": "...", "code": "..."}}
        if isinstance(content.get("error"), dict):
            content["error"]["message"] = message
    else:
        # OpenAI-compatible: {"error": {"message": "...", "type": "...", ...}}
        if isinstance(content.get("error"), dict):
            content["error"]["message"] = message


def register_error_handlers(app: FastAPI) -> None:
    """Register provider-aware error handlers on *app*."""
    app.add_exception_handler(HTTPException, _http_exception_handler)  # type: ignore[arg-type]
    app.add_exception_handler(RequestValidationError, _validation_exception_handler)  # type: ignore[arg-type]



"""Chaos Engineering middleware for LLMock."""

from __future__ import annotations

import asyncio
import os
import random

from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware

from llmock.simulation import (
    ERROR_RATE_ENV_PREFIX,
    SUPPORTED_ERROR_STATUS_CODES,
    build_error_response,
)


class ChaosSettings:
    """Runtime settings for latency and injected HTTP errors."""

    def __init__(
        self,
        *,
        latency_ms: int = 0,
        error_rates: dict[int, float] | None = None,
        error_rate_429: float = 0.0,
        error_rate_500: float = 0.0,
        error_rate_503: float = 0.0,
        **named_error_rates: float,
    ) -> None:
        object.__setattr__(self, "latency_ms", latency_ms)
        object.__setattr__(self, "_error_rates", {})

        combined = dict(error_rates or {})
        if error_rate_429 > 0.0:
            combined[429] = error_rate_429
        if error_rate_500 > 0.0:
            combined[500] = error_rate_500
        if error_rate_503 > 0.0:
            combined[503] = error_rate_503

        for key, value in named_error_rates.items():
            status_code = self._status_code_from_attr(key)
            if status_code is None:
                raise TypeError(f"Unexpected chaos setting '{key}'.")
            combined[status_code] = value

        self.error_rates = combined

    def __getattr__(self, name: str) -> float:
        status_code = self._status_code_from_attr(name)
        if status_code is not None:
            return self._error_rates.get(status_code, 0.0)
        raise AttributeError(name)

    def __setattr__(self, name: str, value: int | float) -> None:
        status_code = self._status_code_from_attr(name)
        if status_code is not None:
            self._error_rates[status_code] = float(value)
            return
        object.__setattr__(self, name, value)

    @property
    def error_rates(self) -> dict[int, float]:
        return dict(self._error_rates)

    @error_rates.setter
    def error_rates(self, rates: dict[int, float]) -> None:
        normalized: dict[int, float] = {}
        for status, rate in rates.items():
            normalized[int(status)] = float(rate)
        object.__setattr__(self, "_error_rates", normalized)

    @staticmethod
    def _status_code_from_attr(name: str) -> int | None:
        if not name.startswith("error_rate_"):
            return None
        suffix = name.removeprefix("error_rate_")
        if not suffix.isdigit():
            return None
        return int(suffix)

    @classmethod
    def from_env(cls) -> "ChaosSettings":
        parsed_rates: dict[int, float] = {}
        for key, value in os.environ.items():
            if not key.startswith(ERROR_RATE_ENV_PREFIX):
                continue
            suffix = key[len(ERROR_RATE_ENV_PREFIX) :]
            if not suffix.isdigit():
                continue
            parsed_rates[int(suffix)] = float(value)
        return cls(
            latency_ms=int(os.getenv("LLMOCK_LATENCY_MS", "0")),
            error_rates=parsed_rates,
        ).validated()

    def with_overrides(
        self,
        *,
        latency_ms: int | None = None,
        error_rates: dict[int, float] | None = None,
        error_rate_429: float | None = None,
        error_rate_500: float | None = None,
        error_rate_503: float | None = None,
    ) -> "ChaosSettings":
        combined = dict(self.error_rates)
        if error_rates:
            combined.update(error_rates)
        if error_rate_429 is not None:
            combined[429] = error_rate_429
        if error_rate_500 is not None:
            combined[500] = error_rate_500
        if error_rate_503 is not None:
            combined[503] = error_rate_503
        return type(self)(
            latency_ms=self.latency_ms if latency_ms is None else latency_ms,
            error_rates=combined,
        ).validated()

    def validated(self) -> "ChaosSettings":
        if self.latency_ms < 0:
            raise ValueError("LLMOCK_LATENCY_MS must be greater than or equal to 0.")

        total_probability = 0.0
        for status, rate in self.error_rates.items():
            if status not in SUPPORTED_ERROR_STATUS_CODES:
                raise ValueError(
                    f"LLMOCK_ERROR_RATE_{status} must target an HTTP error code between 400 and 599."
                )
            if not 0.0 <= rate <= 1.0:
                raise ValueError(f"LLMOCK_ERROR_RATE_{status} must be between 0.0 and 1.0.")
            total_probability += rate

        if total_probability > 1.0:
            raise ValueError("The sum of simulated error probabilities must be <= 1.0.")

        return self

    def as_env(self) -> dict[str, str]:
        env = {"LLMOCK_LATENCY_MS": str(self.latency_ms)}
        for status, rate in self.error_rates.items():
            env[f"{ERROR_RATE_ENV_PREFIX}{status}"] = str(rate)
        return env


chaos_settings = ChaosSettings.from_env()


class ChaosMiddleware(BaseHTTPMiddleware):
    """Middleware that injects latency and random errors based on ChaosSettings."""

    def __init__(self, app, settings: ChaosSettings | None = None) -> None:
        super().__init__(app)
        self.settings = (settings or ChaosSettings.from_env()).validated()

    async def dispatch(self, request: Request, call_next):
        cfg = self.settings

        if request.url.path == "/health":
            return await call_next(request)

        forced_status = request.headers.get("x-llmock-force-status")
        if forced_status:
            status_code = int(forced_status)
            if 400 <= status_code < 600:
                return build_error_response(request.url.path, status_code)

        if cfg.latency_ms > 0:
            await asyncio.sleep(cfg.latency_ms / 1000.0)

        sampled_status = _sample_error_status(cfg)
        if sampled_status is not None:
            return build_error_response(request.url.path, sampled_status)

        return await call_next(request)


def _sample_error_status(settings: ChaosSettings) -> int | None:
    r = random.random()
    cumulative = 0.0
    for status, rate in sorted(settings.error_rates.items()):
        cumulative += rate
        if r < cumulative:
            return status
    return None



"""Tests for provider-specific simulated error payloads."""

from fastapi.testclient import TestClient

from llmock.chaos import ChaosSettings
from llmock.main import create_app


def test_openai_error_shape():
    client = TestClient(create_app(chaos=ChaosSettings(error_rate_429=1.0)))
    response = client.get("/v1/models")
    assert response.status_code == 429
    body = response.json()
    assert body["error"]["type"] == "rate_limit_error"
    assert body["error"]["code"] == "rate_limit_exceeded"


def test_anthropic_error_shape():
    client = TestClient(create_app(chaos=ChaosSettings(error_rate_429=1.0)))
    response = client.get("/anthropic/v1/models")
    assert response.status_code == 429
    body = response.json()
    assert body["type"] == "error"
    assert body["error"]["type"] == "rate_limit_error"


def test_cohere_error_shape():
    client = TestClient(create_app(chaos=ChaosSettings(error_rates={401: 1.0})))
    response = client.get("/cohere/v2/models")
    assert response.status_code == 401
    body = response.json()
    assert body["type"] == "authentication_error"
    assert "message" in body


def test_gemini_error_shape():
    client = TestClient(create_app(chaos=ChaosSettings(error_rate_429=1.0)))
    response = client.get("/gemini/v1beta/models")
    assert response.status_code == 429
    body = response.json()
    assert body["error"]["status"] == "RESOURCE_EXHAUSTED"
    assert body["error"]["code"] == 429



"""Router registry for LLMock.

Provider modules call :func:`register` with their ``APIRouter`` instance(s)
at import time.  :func:`create_app` then calls :func:`get_all_routers` once
to mount every provider in a single loop — no manual edits to ``main.py``
are ever needed when adding a new provider.

Example (in a new provider module)::

    from fastapi import APIRouter
    from llmock.routers import registry

    router = APIRouter(prefix="/myprovider/v1", tags=["myprovider"])

    @router.post("/chat/completions")
    def chat(...): ...

    # Register at module level so create_app() picks this up automatically.
    registry.register(router)
"""

from __future__ import annotations

from fastapi import APIRouter

_routers: list[APIRouter] = []


def register(*routers: APIRouter) -> None:
    """Add one or more :class:`~fastapi.APIRouter` instances to the registry."""
    _routers.extend(routers)


def get_all_routers() -> list[APIRouter]:
    """Return a snapshot of all registered :class:`~fastapi.APIRouter` instances."""
    return list(_routers)
